/**
  *  \file client/dialogs/histselect.cc
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/histselect.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/histlist.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/commands.h"
#include "game/pref.h"
#include "ui/widgets/quit.h"

static ConfigIntOption opt_sort_history(getUserPreferences(), "Sort.History", ValueIntParser::instance);

int
chooseHistoryShip(GPoint pt, int currentShipId)
{
    /* Initial configuration */
    WHistoryShipSelection config;
    config.location = pt;
    switch (int mode = opt_sort_history()) {
     case WHistoryShipSelection::ById:
     case WHistoryShipSelection::ByOwner:
     case WHistoryShipSelection::ByHull:
     case WHistoryShipSelection::ByAge:
     case WHistoryShipSelection::ByName:
        config.sortOrder = static_cast<WHistoryShipSelection::Sort>(mode);
        break;
     default:
        config.sortOrder = WHistoryShipSelection::ById;
        break;
    }
    config.mode = config.getInitialMode(getDisplayedTurn().getCurrentUniverse());

    /* Refuse if we have a position, but the initial mode is not LocalShips.
       This means there are no ships at the requested position. */
    if ((pt.x > 0 || pt.y > 0) && (config.mode != config.LocalShips && config.mode != config.ExactShips)) {
        return 0;
    }

    /* Initial list */
    GObjectList list;
    config.buildList(list, getDisplayedTurn().getCurrentUniverse());
    if (list.size() == 0) {
        return 0;
    }

    /* Build dialog:
       VBox
         HBox
           WHistoryShipList
           UIScrollbar
         HBox
           "OK"
           "Cancel"
           WHistoryShipControl
           UISpacer
           "Help" */
    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Select Ship"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);

    UIFrameGroup& g1 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    WHistoryShipList& list_widget = h.add(new WHistoryShipList(0, 15, font_ems[FONT_NORMAL]*30, list, config));
    g1.add(list_widget);
    g1.add(h.add(new UIScrollbar(list_widget)));
    win.add(g1);

    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g2.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    g2.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    WHistoryShipControl& ctl_widget = h.add(new WHistoryShipControl(config));
    g2.add(ctl_widget);
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(_("Help"), 'h')));
    win.add(g2);
    win.add(h.add(new WHelpWidget("pcc2:historyselect")));
    win.add(h.add(new UIQuit(cm_Escape)));
    if (currentShipId != 0) {
        list_widget.selectObject(GObjectReference(current_histship_type, currentShipId));
    }

    ctl_widget.sig_changed.add(&list_widget, &WHistoryShipList::updateList);

    /* Do it */
    win.pack();
    ui_root->centerWidget(win);
    int val = ui_root->runModal(win);
    opt_sort_history.set(config.sortOrder);
    if (val == cm_Enter) {
        return list_widget.getCurrentObject()->getId();
    } else {
        return 0;
    }
}
