/**
  *  \file client/scr-result.cc
  *  \brief Battle Result Screen
  *
  *  This contaisn the routines for starting a simulation and
  *  displaying its result summary.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/scr-result.h"

#include "client/vcrplay.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/event.h"
#include "ui/layout.h"
#include "util/misc.h"
#include "util/nls.h"
#include "ui/widget-busy.h"
#include "client/scr-sim.h"
#include "ui/table.h"
#include "defs.h"
#include "game/playerset.h"
#include "game/racename.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/button.h"
#include "client/widgets/helpwidget.h"
#include "game/classicvcr.h"
#include "client/widgets/simlist.h"
#include "ui/richdocument.h"
#include "ui/util.h"
#include "ui/widgets/quit.h"

#ifdef CONF_FLAK_SUPPORT
# include "flak/flak-db.h"
# include "flak/vis-ui.h"
#endif

/******************************** Playback *******************************/

// FIXME! This is almost cut&paste from playvcr.cc. It should probably be moved into a library.

/** Play VCRs. */
static bool
playVcrs(GClassicVcrDatabase& db, WVcrSelectorWindowGameInterface& gameif)
{
    ObjectHolder holder;
    WVcrSelector sel(db);
    WVcrSelectorWindow w(sel, gameif);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(holder.add(new UIButton("\xE2\x86\x91", '-')));
    but.add(holder.add(new UIButton("\xE2\x86\x93", '+')));
    but.add(holder.add(new UIButton(_("ENTER - Play"), SDLK_RETURN)));
    but.add(holder.add(new UISpacer()));
    but.add(holder.add(new UIButton(_("ESC - Exit"), SDLK_ESCAPE)));

    w.setLayout(UIVBoxLayout::instance5);
    w.add(sel);
    w.add(but);
    w.pack();
    ui_root->centerWidget(w);
    ui_root->add(w);
    bool result = (ui_root->run() == cm_Enter);
    ui_root->close(w);
    return result;
}

/************************** WSimClassResultList **************************/

WSimClassResultList::WSimClassResultList(GPlayerSet players, GSimResultSummary& result)
    : UIListbox(0, 15, 40 * font_ems[FONT_NORMAL], 0),
      result(result),
      players(players),
      desc_width(0),
      column_width(0)
{
    setHeaderHeight(getItemHeight());
}

WSimClassResultList::~WSimClassResultList()
{ }

void
WSimClassResultList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    for (int i = from; i < to; ++i) {
        const int me = getTop() + i;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (me < result.getNumClassResults()) {
            const GSimResultSummary::ClassResult& class_result = result.getClassResult(me);
            ctx.setTextAlign(0, 0);
            if (result.getTotalWeight() == 1) {
                outTextF(ctx, area.splitX(desc_width),
                         format("%d (%.1f%%)",
                                numToString(class_result.cumulative_weight),
                                100.0*class_result.cumulative_weight / result.getCumulativeWeight()));
            } else {
                outTextF(ctx, area.splitX(desc_width),
                         format("%.1f%%",
                                100.0*class_result.cumulative_weight / result.getCumulativeWeight()));
            }

            ctx.setTextAlign(2, 0);
            for (int i = 1; i <= NUM_OWNERS; ++i) {
                if (players.contains(i)) {
                    GfxRect item = area.splitX(column_width);
                    if (class_result.owned_units[i] == 0)
                        outTextF(ctx, item, "-");
                    else
                        outTextF(ctx, item, numToString(class_result.owned_units[i]));
                }
            }
        }
    }
}

void
WSimClassResultList::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();

    int x = extent.x + desc_width;
    for (int i = 1; i <= NUM_OWNERS; ++i) {
        if (players.contains(i)) {
            outTextF(ctx, x, extent.y, column_width, player_racenames.getAdjName(i));
            x += column_width;
        }
    }

    UIListbox::drawContent(can);
}

void
WSimClassResultList::onResize(GfxRect old_size, bool by_layout)
{
    UIListbox::onResize(old_size, by_layout);

    // count number of players
    int num_players = 0;
    for (int i = 1; i <= NUM_OWNERS; ++i)
        if (players.contains(i))
            ++num_players;

    // compute sizes
    if (num_players != 0) {
        int raw_desc_width = fonts[FONT_NORMAL]->getTextWidth("999x (100.0%)") + 10;
        if (raw_desc_width > extent.w)
            raw_desc_width = extent.w;

        int raw_col_width = (extent.w - raw_desc_width) / num_players;
        if (raw_col_width < 1)
            raw_col_width = 1;
        if (raw_col_width > font_ems[FONT_NORMAL] * 7)
            raw_col_width = font_ems[FONT_NORMAL] * 7;

        column_width = raw_col_width;
        desc_width   = raw_desc_width; // extent.w - raw_col_width * num_players;
    }
}

void
WSimClassResultList::updateFromResult()
{
    setSize(result.getNumClassResults());
    adjustPosition();
    drawWidget();
    sig_moved.raise(getId());
}

/************************* WSimUnitResultSummary *************************/

class WSimUnitResultSummary : public UIWidget {
    GSimResultSummary& summary;
    WSimList& list;
    SignalConnection conn_scroll;
 public:
    WSimUnitResultSummary(GSimResultSummary& summary, WSimList& list);
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
    void onScroll();
};

WSimUnitResultSummary::WSimUnitResultSummary(GSimResultSummary& summary, WSimList& list)
    : UIWidget(),
      summary(summary),
      list(list),
      conn_scroll(list.sig_moved.add(this, &WSimUnitResultSummary::onScroll))
{ }

void
WSimUnitResultSummary::drawContent(GfxCanvas& can)
{
    const GSimResultSummary::UnitResult& res = summary.getUnitResult(list.getCurrent());

    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getExtent(), FONT_NORMAL);
    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Static, _("Fought:"));
    tw.writeVariableCell(UIColor::tc_Static, _("Survived:"));
    tw.writeVariableCell(UIColor::tc_Static, _("Captured:"));

    tw.newRColumn(font_ems[FONT_NORMAL]*6);
    tw.writeVariableCell(UIColor::tc_Static, numToString(res.fought));
    tw.writeVariableCell(UIColor::tc_Static, numToString(res.won));
    tw.writeVariableCell(UIColor::tc_Static, numToString(res.captured));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Static, format(" (%.1f%%)", 100.0 * res.fought / summary.getCumulativeWeight()));
    tw.writeVariableCell(UIColor::tc_Static, format(" (%.1f%%)", 100.0 * res.won / summary.getCumulativeWeight()));
    tw.writeVariableCell(UIColor::tc_Static, format(" (%.1f%%)", 100.0 * res.captured / summary.getCumulativeWeight()));

    tw.finish();
    can.update(getExtent());
}

void
WSimUnitResultSummary::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = GfxPoint(15 * font_ems[FONT_NORMAL], 3 * font_heights[FONT_NORMAL]);
    info.flags = li_Flexible;
}

void
WSimUnitResultSummary::onScroll()
{
    drawWidget();
}

/************************* WSimUnitStat *************************/

class WSimUnitStat : public UIRichDocument {
    // Link-to-item mapping:
    enum LinkItem {
        TorpsFiredLink,
        FightersLostLink,
        DamageTakenLink,
        ShieldsLeftLink,
        CrewLeftOrDefenseLostLink,
        TorpsHitLink,
        MinFightersAboardLink
    };

    GSimResultSummary& summary;
    WSimList& list;
    const GSimState& initial_state;
    SignalConnection conn_scroll;

    void render();
    void onLinkClick(string_t str);
 public:
    WSimUnitStat(GSimResultSummary& summary, WSimList& list, const GSimState& initial_state);
    void onScroll();

    Signal1<Ptr<GVcrDatabase> > sig_watch_vcr;
};

WSimUnitStat::WSimUnitStat(GSimResultSummary& summary, WSimList& list, const GSimState& initial_state)
    : UIRichDocument(GfxPoint(font_sizes[FONT_NORMAL].scaledBy(20, 6)), 0),
      summary(summary),
      list(list),
      initial_state(initial_state),
      conn_scroll(list.sig_moved.add(this, &WSimUnitStat::onScroll)),
      sig_watch_vcr()
{
    sig_link_click.add(this, &WSimUnitStat::onLinkClick);
}

void
WSimUnitStat::onScroll()
{
    render();
    drawWidget();
}

void
WSimUnitStat::render()
{
    if (list.getCurrent() >= summary.getNumUnitResults())
        return;

    const GSimResultSummary::UnitResult& res = summary.getUnitResult(list.getCurrent());

    static const char* const columns[] = {
        N_("Damage:"),
        N_("Shields:"),
        N_("Defense Lost:"),
        N_("SB Ftrs Lost:"),
        N_("Crew Left:"),
        N_("Fighters Lost:"),
        N_("Fighters Left:"),
        N_("Min Ftr aboard:"),
        N_("Torps Launched:"),
        N_("Torps Left:"),
        N_("Torps Hit:"),
        N_("Min Ftrs:"),
    };
    enum { Damage, Shields,
           DefenseLost, SBFtrsLost,
           CrewLeft,
           FightersLost, FightersLeft, MinFtrAboard,
           TorpsLaunched, TorpsLeft, TorpsHit,
           MinSBFtrs };

    struct Link {
        int slot;
        LinkItem item;
        bool reverse;
        Link(int slot, LinkItem item, bool reverse)
            : slot(slot), item(item), reverse(reverse)
            { }
        string_t render(bool max) const
            {
                return format("%d,%d,%d", slot, int(item), int(reverse ^ max));
            }
    };

    struct Helper {
        const int32_t cumulativeWeight;
        GfxFont& font;
        const int fontEm;
        const char*const minLabel;
        const char*const aveLabel;
        const char*const maxLabel;
        const int labelColumnWidth;
        const int minX;
        const int aveX;
        const int maxX;

        Helper(int32_t cumulativeWeight)
            : cumulativeWeight(cumulativeWeight),
              font(*fonts[FONT_NORMAL]),
              fontEm(font_ems[FONT_NORMAL]),
              minLabel(_("min.")),
              aveLabel(_("average")),
              maxLabel(_("max.")),
              labelColumnWidth(getTextWidth(font, StaticTranslatedStringTable(columns, countof(columns)))),
              minX(labelColumnWidth + std::max(font.getTextWidth(minLabel) + 5, 3*fontEm)),
              aveX(minX + std::max(font.getTextWidth(aveLabel) + 5, 4*fontEm)),
              maxX(aveX + std::max(font.getTextWidth(maxLabel) + 5, 3*fontEm))
            { }

        void renderHeadings(RichDocument& doc)
            {
                doc.addRight(minX, RichText(minLabel).withAttribute(RichTextStyleAttribute::Underline));
                doc.addRight(aveX, RichText(aveLabel).withAttribute(RichTextStyleAttribute::Underline));
                doc.addRight(maxX, RichText(maxLabel).withAttribute(RichTextStyleAttribute::Underline));
                doc.addNewline();
            }

        void renderLink(RichDocument& doc, int x, int32_t value, const Ptr<GVcrDatabase>& specimen, const string_t link)
            {
                RichText text(numToString(value));
                if (specimen.ptr() != 0) {
                    doc.addRight(x, text.withAttribute(new RichTextLinkAttribute(link)));
                } else {
                    doc.addRight(x, text);
                }
            }

        void renderRow(RichDocument& doc, size_t rowNr, const GSimStatItem& item, const Link& link)
            {
                doc.add(_(columns[rowNr]));
                renderLink(doc, minX, item.getMin(), item.getMinSpecimen(), link.render(false));
                doc.addRight(aveX, RichText(format("%.1f", double(item.getTotalScaled()) / cumulativeWeight)));
                renderLink(doc, maxX, item.getMax(), item.getMaxSpecimen(), link.render(true));
                doc.addNewline();
            }
    };

    // Prepare the document
    RichDocument& doc = getDocument();
    doc.clear();
    doc.setPageWidth(getExtent().w);

    Helper h(summary.getCumulativeWeight());
    const int slotNr = list.getCurrent();

    // display headings
    h.renderHeadings(doc);

    // display content
    h.renderRow(doc, Damage,  res.damage_taken, Link(slotNr, DamageTakenLink, false));
    h.renderRow(doc, Shields, res.shields_left, Link(slotNr, ShieldsLeftLink, false));

    if (summary.isPlanet(list.getCurrent())) {
        h.renderRow(doc, DefenseLost, res.crew_left_or_defense_lost, Link(slotNr, CrewLeftOrDefenseLostLink, false));
        h.renderRow(doc, SBFtrsLost,  res.fighters_lost,             Link(slotNr, FightersLostLink,          false));
        if (res.fought) {
            // min_fighters_aboard is not valid if the planet did not fight
            // xref GSimState::initializeStats
            h.renderRow(doc, MinSBFtrs, res.min_fighters_aboard,     Link(slotNr, MinFightersAboardLink,     false));
        }
    } else {
        const GSimShip& ship = initial_state.getShip(list.getCurrent());
        h.renderRow(doc, CrewLeft, res.crew_left_or_defense_lost, Link(slotNr, CrewLeftOrDefenseLostLink, false));
        if (ship.getBayCount() > 0) {
            h.renderRow(doc, FightersLost, res.fighters_lost, Link(slotNr, FightersLostLink, false));
            h.renderRow(doc, FightersLeft, GSimStatItem(res.fighters_lost, ship.getAmmo(), summary.getCumulativeWeight()), Link(slotNr, FightersLostLink, true));
            h.renderRow(doc, MinFtrAboard, res.min_fighters_aboard, Link(slotNr, MinFightersAboardLink, false));
        } else if (ship.getTorpLauncherCount() > 0) {
            h.renderRow(doc, TorpsLaunched, res.torps_fired, Link(slotNr, TorpsFiredLink, false));
            h.renderRow(doc, TorpsLeft, GSimStatItem(res.torps_fired, ship.getAmmo(), summary.getCumulativeWeight()), Link(slotNr, TorpsFiredLink, true));
            h.renderRow(doc, TorpsHit, res.torps_hit, Link(slotNr, TorpsHitLink, false));
        }
    }

    // Finish
    doc.finish();
    handleDocumentUpdate();
}

void
WSimUnitStat::onLinkClick(string_t link)
{
    int slot, item, max;
    string_t slotStr, itemStr;
    if (strSplit(link, slotStr, link, ",")
        && strSplit(link, itemStr, link, ",")
        && stringValue(slotStr, slot)
        && stringValue(itemStr, item)
        && stringValue(link, max)
        && slot >= 0
        && slot < int(summary.getNumUnitResults()))
    {
        // Valid link. Fetch the result.
        const GSimResultSummary::UnitResult& unitResult = summary.getUnitResult(slot);

        // Fetch the GSimStatItem
        const GSimStatItem& statItem =
            (item == TorpsFiredLink ? unitResult.torps_fired :
             item == FightersLostLink ? unitResult.fighters_lost :
             item == DamageTakenLink ? unitResult.damage_taken :
             item == ShieldsLeftLink ? unitResult.shields_left :
             item == CrewLeftOrDefenseLostLink ? unitResult.crew_left_or_defense_lost :
             item == TorpsHitLink ? unitResult.torps_hit :
             /* item == MinFightersAboardLink ? */ unitResult.min_fighters_aboard);

        // Fetch the fight
        Ptr<GVcrDatabase> db = max ? statItem.getMaxSpecimen() : statItem.getMinSpecimen();
        if (db) {
            sig_watch_vcr.raise(db);
        }
    }
}

/**************************** WSimResultWindow ***************************/

static GPlayerSet
countPlayers(const GSimState& st)
{
    GPlayerSet result;
    if (st.hasPlanet())
        result += st.getPlanet().getOwner();
    for (GSimState::ship_index_t i = 0; i < st.getNumShips(); ++i)
        result += st.getShip(i).getOwner();
    return result;
}

/** Add result buttons to widget structure.
    \param out Group to add to
    \param holder Space for new object
    \param default_button The card-specific button to use */
static void
addResultButtons(UIGroup& out, ObjectHolder& holder, UIButton& default_button)
{
    UIGroup& group_buttons = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group_buttons2 = holder.add(new UIGroup(UIHBoxLayout::instance5));

    group_buttons.add(default_button);
    group_buttons.add(holder.add(new UIButton(cm_Escape, _("ESC"), SDLK_ESCAPE)).addStop());
    group_buttons.add(holder.add(new UISpacer()));
    group_buttons.add(holder.add(new UIButton(_("Help"), 'h')));

    group_buttons2.add(holder.add(new UIButton(_("Space - Again"), ' ')));
    group_buttons2.add(holder.add(new UIButton(_("S - Series"), 's')));
    group_buttons2.add(holder.add(new UIButton(_("R - Repeat"), 'r')));
    group_buttons2.add(holder.add(new UISpacer()));

    out.add(group_buttons2);
    out.add(group_buttons);
}

/** Constructor. */
WSimResultWindow::WSimResultWindow(GSimulatorGameInterface& game_if, const GSimState& sim_setup, const GSimOptions& sim_options)
    : UIWindow(_("Simulation Results"), BLUE_WINDOW, COLOR_GRAY),
      sim_setup(sim_setup),
      sim_options(sim_options),
      last_result(sim_options),
      result_summary(),
      class_result_list(countPlayers(sim_setup), result_summary),
      summary_line("", UIStaticText::StaticTitle),
      summary_line2("", UIStaticText::StaticTitle),
      tab_content(),
      tabs(tab_content),
      result_list(holder.add(new WSimList(game_if, const_cast<GSimState&>(sim_setup), true))),
      result_list_index(-1)
{
    init();
}

void
WSimResultWindow::init()
{
    /* Widget structure now looks like this:
         Window
           CardGroupTab
           CardGroup
             summary_tab
               list with scrollbar
               summary_line
               button bar
             result_tab
               list, info widget
               summary_line2
               button bar
       That is, we have two copies of the summary_line, and two copies
       of the buttons, although we only want to exchange one button, namely
       'Watch sample' vs 'Edit ship'. Possible alternatives:
       - rearrange stuff. Tried it, but I'm so used to the CCBSim 1.x layout
         that I think the changed one makes it worse.
       - change the label of the button on the fly. Currently not supported
         by the widget system; may make problems with layouting. */
    setLayout(UIVBoxLayout::instance5);

    UIGroup& summary_tab = holder.add(new UIGroup(UIVBoxLayout::instance5, 1000));
    UIGroup& result_tab  = holder.add(new UIGroup(UIVBoxLayout::instance5, 1001));

    UIGroup& summary_content = holder.add(new UIGroup(UIHBoxLayout::instance0));

    summary_content.add(class_result_list);
    summary_content.add(holder.add(new UIScrollbar(class_result_list)));
    summary_tab.add(summary_content);
    summary_tab.add(summary_line);

    UIGroup& result_content = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& result_rhs = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIFrameGroup& result_lhs = holder.add(new UIFrameGroup(UIHBoxLayout::instance0));

    WSimUnitResultSummary& result_sum1 = holder.add(new WSimUnitResultSummary(result_summary, result_list));
    WSimUnitStat& result_sum2 = holder.add(new WSimUnitStat(result_summary, result_list, sim_setup));
    result_lhs.add(result_list);
    result_lhs.add(holder.add(new UIScrollbar(result_list)));
    result_content.add(result_lhs);
    result_content.add(result_rhs);
    result_rhs.add(result_sum1);
    result_rhs.add(result_sum2);
    result_rhs.add(holder.add(new UISpacer()));

    result_tab.add(result_content);
    result_tab.add(summary_line2);

    addResultButtons(summary_tab, holder, holder.add(new UIButton(_("ENTER - Watch sample"), SDLK_RETURN)));
    addResultButtons(result_tab,  holder, holder.add(new UIButton(_("E - Edit this ship"), 'e')));

    tab_content.add(summary_tab);
    tab_content.add(result_tab);

    tabs.addTab(1000, UIKeyString(_("Totals")));
    tabs.addTab(1001, UIKeyString(_("Details")));

    add(tabs);
    add(tab_content);
    add(holder.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Horiz)));
    add(holder.add(new WHelpWidget("pcc2:simresult")));
    add(holder.add(new UIQuit(cm_Escape)));
    class_result_list.setFocusedWidget();

    pack();

    sig_run.add(&class_result_list, &WSimClassResultList::updateFromResult);
    sig_run.add(&result_sum1,       &WSimUnitResultSummary::onScroll);
    sig_run.add(&result_sum2,       &WSimUnitStat::onScroll);

    result_sum2.sig_watch_vcr.add(this, &WSimResultWindow::onWatch);
    result_sum2.onScroll();
}

WSimResultWindow::~WSimResultWindow()
{ }

/** Run simulation.
    \param series true if we're simulating a series; stop after one run
    \param max    maximum number of battles to generate, 0 for infinite

    \todo On SMP/Multicore machines, we could fork off some worker
    threads to do the runSimulation(). Currently, this does not work
    because our VCR algorithms may use static storage. */
void
WSimResultWindow::runSimulation(bool series, int max)
{
    ProgressMonitor monitor;
    UIBusyWidget busy_widget;
    ui_root->add(busy_widget);
    ui_root->performDeferredRedraws();
    monitor.sig_tick.add(&busy_widget, &UIBusyWidget::tick);

    long start = getTimer();
    while (1) {
        GSimState new_state(sim_setup);
        last_result.init(sim_options, result_summary.getNumBattles());
        ::runSimulation(new_state, sim_options, last_result, monitor);
        result_summary.addResult(sim_setup, new_state, last_result);

        /* check exit conditions */
        if (max != 0) {
            if (--max == 0)
                break;
        }
        if (series) {
            if (last_result.series_length != 0)
                if ((result_summary.getNumBattles() % last_result.series_length) == 0)
                    break;
        }

        /* poll events */
        UIEvent event;
        do
            getEvent(event, true);
        while (event.type != ev_Key && event.type != ev_Idle);
        if (event.type == ev_Key && getClassOfKey(event.key.code) == KeyNormal)
            break;

        long now = getTimer();
        if (now > start + TIMER_HZ/2) {
            updateResults();
            ui_root->performDeferredRedraws();
            start = now;
        }
    }
}

/** Run first simulation.
    \return true on success, false if setup doesn't give any fights */
bool
WSimResultWindow::runFirstSimulation()
{
    ProgressMonitor monitor;

    GSimState new_state(sim_setup);
    last_result.init(sim_options, result_summary.getNumBattles());
    ::runSimulation(new_state, sim_options, last_result, monitor);
    if (last_result.battles != 0 && last_result.battles->getNumBattles() != 0) {
        result_summary.addResult(sim_setup, new_state, last_result);
        updateResults();
        return true;
    } else {
        return false;
    }
}

/** Update our widgets from the battle result. */
void
WSimResultWindow::updateResults()
{
    sig_run.raise();

    string_t result = format(_("%d simulation%!1{s%}"), numToString(result_summary.getNumBattles()));
    summary_line.setText(result);
    summary_line2.setText(result);
}

bool
WSimResultWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_RETURN:
            if (tabs.getCurrentTab() == 1000) {
                onWatch(result_summary.getClassResult(class_result_list.getCurrent()).sample_battle);
            }
            return true;
         case 'e':
            if (tabs.getCurrentTab() == 1001) {
                result_list_index = result_list.getCurrent();
                stop(cm_Enter);
            }
            return true;
         case ' ':
            runSimulation(false, 1);
            updateResults();
            return true;
         case 's':
            runSimulation(true, 0);
            updateResults();
            return true;
         case 'r':
            runSimulation(false, 0);
            updateResults();
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

int
WSimResultWindow::getResultListIndex() const
{
    return result_list_index;
}

void
WSimResultWindow::onWatch(Ptr<GVcrDatabase> db)
{
    if (GClassicVcrDatabase* cdb = dynamic_cast<GClassicVcrDatabase*>(db.ptr())) {
        if (playVcrs(*cdb, *this)) {
            stop(cm_Enter);
        }
#ifdef CONF_FLAK_SUPPORT
    } else if (GFlakVcrDatabase* fdb = dynamic_cast<GFlakVcrDatabase*>(db.ptr())) {
        if (fdb->getNumBattles() > 0)
            playFlakBattle(fdb->getBattle(0));
#endif
    }
}

int
WSimResultWindow::findGameObject(const GVcrObject& obj)
{
    if (obj.isPlanet()) {
        if (sim_setup.hasPlanet() && sim_setup.getPlanet().getId() == obj.getId()) {
            return sim_setup.getNumShips();
        } else {
            return -1;
        }
    } else {
        GSimState::ship_index_t idx = sim_setup.getIndexOf(obj.getId());
        if (idx < sim_setup.getNumShips()
            && (sim_setup.getShip(idx).getHull() == 0
                || obj.canBeHull(sim_setup.getShip(idx).getHull())))
        {
            return idx;
        } else {
            return -1;
        }
    }
}

bool
WSimResultWindow::isGameObject(const GVcrObject& obj)
{
    return findGameObject(obj) >= 0;
}

void
WSimResultWindow::goToGameObject(const GVcrObject& obj)
{
    result_list_index = findGameObject(obj);
}
