/**
  *  \file client/widgets/funcinfo.cc
  *  \brief Hull Function Information Widget
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/funcinfo.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/struct.h"
#include "ui/uiglobal.h"
#include "ui/skin.h"
#include "ui/util.h"
#include "game/config.h"

WHullFunctionInfo::WHullFunctionInfo(int owner, int damage)
    : UIRichDocument(font_sizes[FONT_NORMAL].scaledBy(30, 11), 0),
      owner(owner), damage(damage)
{ }

WHullFunctionInfo::~WHullFunctionInfo()
{ }

void
WHullFunctionInfo::setFunction(const GHullFunction* hf)
{
    RichDocument& doc = getDocument();
    doc.clear();
    if (hf != 0) {
        /* Function title */
        const GBasicHullFunction* ptr = GHullFunctionData::instance.getBasicFunctionById(hf->getBasicFunctionId());
        int damageLimit;
        if (ptr) {
            doc.add(RichText(ptr->getDescription()).withAttribute(RichTextStyleAttribute::Bold));
            damageLimit = ptr->getDamageLimit(owner, config);
        } else {
            doc.add(RichText(format(_("Hull Function #%d"), hf->getBasicFunctionId())).withAttribute(RichTextStyleAttribute::Bold));
            damageLimit = -1;
        }
        doc.addParagraph();

        /* Availability information */
        string_t tmp = formatPlayerSet(hf->getPlayers());
        if (tmp.size()) {
            doc.add(string_t(format(_("Available to %s"), tmp)));
            doc.addNewline();
        }
        tmp = formatExpLevelSet(hf->getLevels());
        if (tmp.size()) {
            int i = 0;
            while (i <= NUM_EXPERIENCE_LEVELS && !hf->getLevels().contains(i)) {
                ++i;
            }
            if (i > 0 && i <= NUM_EXPERIENCE_LEVELS) {
                tmp += format(_(" (%d EP)"), config.ExperienceLevels(i));
            }
            doc.add(string_t(format(_("Available at %s"), tmp)));
            doc.addNewline();
        }

        if (hf->getKind() == GHullFunction::AssignedToRace) {
            if (hf->getPlayers().contains(GPlayerSet::allUpTo(NUM_PLAYERS) - 0)) {
                doc.add(_("Universal ability"));
            } else {
                doc.add(_("Racial ability"));
            }
        } else if (hf->getKind() == GHullFunction::AssignedToShip) {
            doc.add(_("Assigned to ship"));
        } else {
            doc.add(_("Assigned to ship class"));
        }

        if (damageLimit >= 0 && damageLimit < 100) {
            doc.addNewline();
            doc.add(string_t(format(_("Fails at %d%% damage"), damageLimit)));
            if (damage >= damageLimit) {
                doc.add(" ");
                doc.add(RichText(UIColor::tc_Red, _("(currently broken)")));
            }
        }

        doc.addParagraph();

        /* Description */
        if (ptr) {
            doc.add(ptr->getExplanation());
        }
    }
    doc.finish();
    handleDocumentUpdate();
}
