/**
  *  \file client/wscanchart.cc
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wscanchart.h"
#include "gfx/clip.h"
#include "gfx/palette.h"
#include "game/lock.h"
#include "game/config.h"
#include "game/host.h"
#include "ui/uiglobal.h"
#include "ui/prefix.h"
#include "ui/util.h"
#include "game/formula.h"
#include "game/racename.h"
#include "game/ship.h"
#include "game/planet.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/team.h"
#include "game/global.h"
#include "client/markerconfig.h"
#include "game/pref.h"
#include "ui/widgets/quit.h"

/************************** WScannerChartWidget **************************/

/*! \class WScannerChartWidget
    \brief Chart Widget for Control Screen Scanners

    This contains a standard chart widget with a movable scanner beam. */

/** Constructor. */
WScannerChartWidget::WScannerChartWidget()
    : scan_origin(getViewport().center),
      scan_target(getViewport().center),
      conn_change(observe(getDisplayedTurn().getCurrentUniverse().sig_universe_changed)),
      conn_options(observe(getChartOpts(true, 1, 1).sig_changed)),
      tt_ctrl(*this)
{ }

WScannerChartWidget::~WScannerChartWidget()
{ }

void
WScannerChartWidget::drawScannerXor(GfxCanvas& can)
{
    GfxXorFilter      xor_flt(can);
    GfxClipRectFilter can1(getExtent(), xor_flt);

    drawScanner(can1);
}

/** Draw scanner beam. */
void
WScannerChartWidget::drawScanner(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_YELLOW);

    getViewport().drawScanner(ctx, scan_origin, scan_target);
}

/** Query current location information for planet lock.
    \param hyperjumping        [out] true if we're a ship that is hyperjumping
    \param current_gravity_pid [out] Id of planet whose warp well we're in */
void
WScannerChartWidget::lockQueryLocation(bool& hyperjumping, int& current_gravity_pid)
{
    hyperjumping        = false;
    current_gravity_pid = (isValid()
                           ? getDisplayedTurn().getCurrentUniverse().getGravityPlanetAt(getScannerOrigin())
                           : 0);
}

/** Query distance to a point. Returns any distance metric. Can return
    -1 to reject a point. */
int32
WScannerChartWidget::lockQueryDistance(GPoint pt)
{
    if (isValid())
        return distanceSquared(getScannerOrigin(), pt);
    else
        return -1;
}

/** Draw below starchart. This draws the center cross. */
void
WScannerChartWidget::drawPre(GfxCanvas& can)
{
    const GfxPoint m = getViewport().scaleXY(getScannerOrigin());

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_BLUEBLACK);

    drawHLine(ctx, m.x - 10, m.y,      m.x + 11);
    drawVLine(ctx, m.x,      m.y - 10, m.y + 11);
}

/** Draw above starchart. This draws the scanner. */
void
WScannerChartWidget::drawPost(GfxCanvas& can)
{
    if (isValid())
        drawScannerXor(can);
}

/** Clip scanner position.
    \return pt modified such that it is in the visible range
    \todo maybe move to WChartWidget? */
GPoint
WScannerChartWidget::clipScannerPosition(GPoint pt) const
{
    const GChartViewport& vp = getViewport();
    const GPoint m           = vp.center;
    const int xw = vp.unscale(extent.w) / 2;
    const int yw = vp.unscale(extent.h) / 2;
    if (pt.x < m.x - xw)
        pt.x = m.x - xw;
    if (pt.x > m.x + xw)
        pt.x = m.x + xw;
    if (pt.y < m.y - yw)
        pt.y = m.y - yw;
    if (pt.y > m.y + yw)
        pt.y = m.y + yw;
    return pt;
}

/** Set scanner target.
    \param pt      point
    \param limit   true to limit actual value to visible range,
                   false to take new value without checking */
void
WScannerChartWidget::setScannerTarget(GPoint pt)
{
    if (pt != scan_target) {
        if (hasState(st_Exposed) && !hasState(st_WantRedraw)) {
            /* Widget is visible and up-to-date. Use efficient shortcut
               instead of redrawing whole widget. */
            ObjectHolder h;
            GfxCanvas& can = getParent()->getCanvasFor(*this, h);
            can.lock();
            drawScannerXor(can);
            scan_target = pt;
            drawScannerXor(can);
            can.unlock();
            can.update(extent);
            sig_moved.raise(getId());
        } else {
            /* Widget is either not visible, or will be redrawn soon,
               so we don't need to do anything */
            scan_target = pt;
            sig_moved.raise(getId());
        }
    }
}

void
WScannerChartWidget::setScannerOrigin(GPoint pt)
{
    if (pt != scan_origin) {
        scan_origin = pt;
        sig_moved.raise(getId());
        drawWidget();
    }
}

void
WScannerChartWidget::userMoveScannerTo(GPoint pt)
{
    setScannerTarget(pt);
}

/** Move according to a keypress. */
void
WScannerChartWidget::moveKey(int dx, int dy)
{
    setSelectedWidget();
    GPoint pt = getScannerTarget();
    userMoveScannerTo(clipScannerPosition(GPoint(pt.x + dx, pt.y + dy)));
}

/** Display tooltip for a given mouse position. */
void
WScannerChartWidget::doTooltip(GfxPoint pt)
{
    // Valid?
    if (!isValid())
        return;

    // Figure out where we clicked
    const GChartViewport& vp = getViewport();
    const GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GLockData locker(vp.unscaleXY(pt),
                     GPoint(vp.unscaleX(extent.x),
                            vp.unscaleY(extent.y + extent.h - 1)),
                     GPoint(vp.unscaleX(extent.x + extent.w - 1),
                            vp.unscaleY(extent.y)),
                     false);
    findShip(univ, locker);
    findPlanet(univ, locker);
    // findMarker(univ, locker); FIXME? Would be nice

    // Distance must be small enough (50 means distance of ~7 pixels; has been pulled out
    // of a hat, and it's the same as used in PCC 1.x)
    const GfxPoint resolved = vp.scaleXY(locker.getPoint());
    if (square(resolved.x - pt.x) + square(resolved.y - pt.y) > 50)
        return;

    // Now check who's there. First the planet:
    const GPoint real_pos = locker.getPoint().getCanonicalLocation();
    string_t desc;
    if (int pid = univ.getPlanetAt(real_pos)) {
        desc += univ.getPlanet(pid).getName(GObject::LongName);
        desc += "\n";
    }

    // Count ships
    GPlayerArray<int> numShips;
    int myShipId = 0;
    numShips.setAll(0);
    for (int sid = univ.ty_any_ships.findNextIndex(0); sid != 0; sid = univ.ty_any_ships.findNextIndex(sid)) {
        const GShip& sh = univ.getShip(sid);
        if (sh.getPos() == real_pos && numShips.isValidIndex(sh.getOwner())) {
            // Valid ship, count it
            ++numShips[sh.getOwner()];
            if (myShipId == 0 && sh.getOwner() == getPlayerId())
                myShipId = sid;
        }
    }

    // Own ships?
    if (myShipId != 0) {
        if (numShips[getPlayerId()] > 1)
            desc += format(_("%s + %d own ship%!1{s%}"), univ.getShip(myShipId).getName(GObject::LongName), numShips[getPlayerId()]-1);
        else
            desc += univ.getShip(myShipId).getName(GObject::LongName);
        desc += "\n";
    }

    // Foreign ships?
    for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
        if (pl != getPlayerId() && numShips[pl] != 0)
            desc += format(_("%d %s ship%!1{s%}\n"), numShips[pl], player_racenames.getAdjName(pl));
    }

    // Display tooltip if there were any units
    if (desc.size() != 0) {
        ::doTooltip(desc.substr(0, desc.size()-1), resolved);
    }
}

/** Handle keypress.
    \param k key
    \return true iff key handled */
bool
WScannerChartWidget::handleKey(UIEvent::Key k)
{
    int n = k.arg;
    switch (k.code) {
     case SDLK_LEFT:
        moveKey(n ? -n : -10, 0);
        return true;
     case SDLK_RIGHT:
        moveKey(n ? n : 10, 0);
        return true;
     case SDLK_UP:
        moveKey(0, n ? n : 10);
        return true;
     case SDLK_DOWN:
        moveKey(0, n ? -n : -10);
        return true;
     case ss_Shift + SDLK_LEFT:
        moveKey(-1, 0);
        return true;
     case ss_Shift + SDLK_RIGHT:
        moveKey(1, 0);
        return true;
     case ss_Shift + SDLK_UP:
        moveKey(0, 1);
        return true;
     case ss_Shift + SDLK_DOWN:
        moveKey(0, -1);
        return true;
     case ss_Ctrl + SDLK_LEFT:
        moveKey(-100, 0);
        return true;
     case ss_Ctrl + SDLK_RIGHT:
        moveKey(100, 0);
        return true;
     case ss_Ctrl + SDLK_UP:
        moveKey(0, 100);
        return true;
     case ss_Ctrl + SDLK_DOWN:
        moveKey(0, -100);
        return true;
     case '+':
        if (hasState(sc_Zoomable))
            zoomIn();
        return true;
     case '-':
        if (hasState(sc_Zoomable))
            zoomOut();
        return true;
     case ' ':                               doItemLock(getScannerTarget(), lm_Left,  false, false); return true;
     case SDLK_RETURN:                       doItemLock(getScannerTarget(), lm_Right, false, false); return true;
     case ss_Ctrl + ' ':                     doItemLock(getScannerTarget(), lm_Left,  true,  false); return true;
     case ss_Ctrl + SDLK_RETURN:             doItemLock(getScannerTarget(), lm_Right, true,  false); return true;
     case ss_Shift + ' ':                    doItemLock(getScannerTarget(), lm_Left,  false, true);  return true;
     case ss_Shift + SDLK_RETURN:            doItemLock(getScannerTarget(), lm_Right, false, true);  return true;
     case ss_Shift + ss_Ctrl + ' ':          doItemLock(getScannerTarget(), lm_Left,  true,  true);  return true;
     case ss_Shift + ss_Ctrl + SDLK_RETURN:  doItemLock(getScannerTarget(), lm_Right, true,  true);  return true;
     case ss_Alt + 'r':
        moveKey(-5 + rand() % 11, -5 + rand() % 11);
        return true;
     case 'm':
     case ss_Ctrl + 'm':
        {
            // FIXME: this duplicates createCannedMarker(). Can we share it?
            ConfigMarkerOption& opt = ConfigMarkerOption::getCannedMarker(k.arg % ConfigMarkerOption::NUM_CANNED_MARKERS);

            // Create marker
            GDrawing marker(getScannerTarget(), MarkerDrawing);
            marker.setMarkerKind(opt.getType());
            marker.setColor(marker.getColorFromUserColor(opt.getColorIndex()));
            getDisplayedTurn().getCurrentUniverse().getDrawings().add(marker);

            // FIXME: tell background to redraw - how?
        }
        drawWidget();
        return true;

        /* z [edit zoom in PCC 1.x] */
    }
    return false;
}

bool
WScannerChartWidget::handleOption(UIEvent::Key k)
{
    if ((k.code & ss_Alt) == ss_Alt) {
        const GChartViewport& vp = getViewport();
        if (getChartOpts(true, vp.mult, vp.divi).toggleOptionKey(k.code & ~(ss_Alt | ss_Ctrl))) {
            return true;
        }
    }

    return false;
}

/** Lock at item. This adjusts the clicked point, and calls userMoveScannerTo()
    with the new position.
    \param clicked       Point clicked by user
    \param mode          Lock mode (left/right mouse button)
    \param marked_only   true to accept only marked units
    \param optimize_warp true to optimize using gravity wells */
void
WScannerChartWidget::doItemLock(GPoint clicked,
                                TLockMode mode,
                                bool marked_only,
                                bool optimize_warp)
{
    // Regular lock code
    const GChartViewport& vp = getViewport();
    const GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GLockData locker(clicked,
                     GPoint(vp.unscaleX(extent.x),
                            vp.unscaleY(extent.y + extent.h - 1)),
                     GPoint(vp.unscaleX(extent.x + extent.w - 1),
                            vp.unscaleY(extent.y)),
                     marked_only);

    if (hasState(sc_ShipsOnly)) {
        // Ship-only (Intercept mission)
        findShip(univ, locker);
        userMoveScannerTo(locker.getPoint());
    } else if ((getLockMode(mode) & li_Planet) != 0 && (optimize_warp ^ !!getUserPreferences().ChartScannerWarpWells())) {
        // Optimize warp wells
        if (optimize_warp) {
            // User explicitly pressed additional key to enable locking, so we can limit search to planets
            findPlanet(univ, locker);
        } else {
            // User did not press additional key, but configured warp-lock as default.
            // Do NOT limit search to planets, to still allow locking on non-planets without additional key.
            findItem(univ, locker, mode);
        }
        GPoint found_point = locker.getPoint();

        // Query current position
        bool hyperjumping;
        int current_pid;
        lockQueryLocation(hyperjumping, current_pid);

        // Can we optimize warp wells?
        int clicked_pid = univ.getPlanetAt(found_point);
        if (clicked_pid > 0
            && config.AllowGravityWells()
            && (!hyperjumping || !host.isPHost() || config.AllowHyperjumpGravWells())
            && clicked_pid != current_pid)
        {
            /* We try to find the edge of a gravity well unless
               - we're heading for deep space, i.e. no planet found
               - gravity wells are disabled
               - we're starting inside the same gravity well we started in,
                 in this case we assume we want to move to the planet */
            const int wwrange = (host.isPHost() ? config.GravityWellRange() :
                                 hyperjumping ? 2 : 3);

            // Start with the assumption that moving directly is the best choice.
            // Then try all points in warp well range.
            int32 best_dist   = lockQueryDistance(found_point);
            GPoint best_point = found_point;
            for (int dx = -wwrange; dx <= wwrange; ++dx) {
                for (int dy = -wwrange; dy <= wwrange; ++dy) {
                    GPoint new_point = GPoint(found_point.x + dx, found_point.y + dy);
                    int32 new_dist = lockQueryDistance(new_point);
                    if (new_dist >= 0
                        && (best_dist < 0 || new_dist < best_dist)
                        && univ.getGravityPlanetAt(new_point) == clicked_pid)
                    {
                        // Accept new point if it is valid, has a better metric than
                        // the previous one, and it is in the same warp well.
                        best_dist = new_dist;
                        best_point = new_point;
                    }
                }
            }

            // Move to found point, if any
            if (best_dist >= 0)
                userMoveScannerTo(best_point);
        } else {
            // No warp wells, so just move to found point
            userMoveScannerTo(found_point);
        }
    } else {
        // Regular locking only
        findItem(univ, locker, mode);
        userMoveScannerTo(locker.getPoint());
    }
}

bool
WScannerChartWidget::handleEvent(const UIEvent& e, bool second_pass)
{
    /* nothing selected -> no input accepted */
    if (!isValid())
        return WChartWidget::handleEvent(e, second_pass);

    if (tt_ctrl.checkEvent(e))
        doTooltip(tt_ctrl.getTooltipPosition());

    if (e.type == ev_Mouse && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        const GChartViewport& vp = getViewport();
        setSelectedWidget();
        if (e.mouse.b || hasState(sc_KeyboardMode)) {
            if (e.mouse.b & mb_Double) {
                sig_dblclick.raise(getId());
            } else {
                if (e.mouse.b & mb_Left)
                    doItemLock(vp.unscaleXY(GfxPoint(e.mouse.x, e.mouse.y)),
                               lm_Left,
                               (e.mouse.b & ss_Ctrl) != 0,
                               (e.mouse.b & ss_Shift) != 0);
                else if (e.mouse.b & mb_Right)
                    doItemLock(vp.unscaleXY(GfxPoint(e.mouse.x, e.mouse.y)),
                               lm_Right,
                               (e.mouse.b & ss_Ctrl) != 0,
                               (e.mouse.b & ss_Shift) != 0);
                else
                    userMoveScannerTo(vp.unscaleXY(GfxPoint(e.mouse.x, e.mouse.y)));
            }
        }
        return true;
    } else if (e.type == ev_Key && hasState(sc_KeyboardMode | sc_StyleChange) && handleOption(e.key)) {
        return true;
    } else if (e.type == ev_Key && hasState(sc_KeyboardMode | sc_Keyboard)) {
        return handleKey(e.key);
    } else {
        return WChartWidget::handleEvent(e, second_pass);
    }
}

void
WScannerChartWidget::onStateChange(int state, bool enable)
{
    tt_ctrl.checkState();
    WChartWidget::onStateChange(state, enable);
}

/***************************** Keyboard Mode *****************************/

static inline void
drawActivityFrame(GfxCanvas& can, const GfxRect& r, uint8 p)
{
    drawDashedFrame(can, r, p, standard_colors[COLOR_YELLOW], standard_colors[COLOR_DARK]);
}

namespace {
    class KeyboardModeHelper {
        WScannerChartWidget& w;
     public:
        KeyboardModeHelper(WScannerChartWidget& w);
        ~KeyboardModeHelper();

        void drawFrame(uint8 phase);
        void clearFrame();
    };
}

KeyboardModeHelper::KeyboardModeHelper(WScannerChartWidget& w)
    : w(w)
{
    w.setState(w.sc_Keyboard, true);
    w.setSelectedWidget();
}

KeyboardModeHelper::~KeyboardModeHelper()
{
    w.setState(w.sc_Keyboard, false);
    clearFrame();
}

void
KeyboardModeHelper::drawFrame(uint8 phase)
{
    // Obtain canvas and size
    ObjectHolder tmp;
    GfxCanvas& can = w.getParent()->getCanvasFor(w, tmp);
    GfxRect frame_size = w.getExtent();
    frame_size.grow(2, 2);

    // Draw
    can.lock();
    drawActivityFrame(can, frame_size, phase);
    can.unlock();
    can.update(frame_size);
}

void
KeyboardModeHelper::clearFrame()
{
    GfxRect frame_size = w.getExtent();
    frame_size.grow(2, 2);
    ui_root->markRedrawRegion(frame_size);
}

/** Perform keyboard movement mode. This gives the WScannerChartWidget
    exclusive focus, and lets the user move the pointer using the
    keyboard or the mouse. This function returns when the user exits
    keyboard mode. */
void
WScannerChartWidget::doKeyboardMode()
{
    // Keyboard mode not available when there's no valid position
    if (!isValid())
        return;

    // Avoid recursion
    if (hasState(sc_Keyboard))
        return;

    // Regular keyboard mode
    KeyboardModeHelper helper(*this);
    uint8 frame_phase = 0;
    bool need_update_frame = true;
    bool running = true;
    int pending_dx = 0;
    int pending_dy = 0;
    while (running) {
        // Perform updates
        ui_root->performDeferredRedraws();
        if (need_update_frame) {
            helper.drawFrame(frame_phase/4);
            need_update_frame = false;
        }

        // Handle events
        UIEvent e;
        getEvent(e, false, true);
        switch (e.type) {
         case ev_Tick:
            // Slow down a little
            if ((++frame_phase & 3) == 0)
                need_update_frame = true;
            break;
         case ev_Key:
            /* PCC 1's exit commands:
                 1..9         prefix arg
                 BS, ESC, y   regular exit
                 x            exit + 'x'
                 C-w          exit + C-w
                 l            exit + 'l'
               We add some more. */
            switch (e.key.code) {
             case SDLK_ESCAPE:
             case SDLK_BACKSPACE:
             case 'y':
             case vk_Quit:
                running = false;
                break;

             case SDLK_F1 + ss_Ctrl:
             case SDLK_F2 + ss_Ctrl:
             case SDLK_F3 + ss_Ctrl:
             case SDLK_F4 + ss_Ctrl:
             case SDLK_F5 + ss_Ctrl:
             case 'x':
             case 'w' + ss_Ctrl:
             case 'L':
             case 'l':
                running = false;
                ungetEvent(e);
                break;
             default:
                if (e.key.code >= '1' && e.key.code <= '9') {
                    startPrefixArg(e.key.code - '0');
                } else {
                    if (!handleOption(e.key)) {
                        handleKey(e.key);
                    }
                }
                break;
            }
            break;
         case ev_Mouse:
            {
                // Perform relative mouse movement
                pending_dx += e.mouse.x;
                pending_dy += e.mouse.y;
                GPoint new_pt = GPoint(scan_target.x + getViewport().unscale(pending_dx),
                                       scan_target.y - getViewport().unscale(pending_dy));
                if (new_pt != scan_target)
                    pending_dx = pending_dy = 0;

                // Find new position, by locking if needed
                new_pt = clipScannerPosition(new_pt);
                if (e.mouse.b & mb_Left) {
                    doItemLock(new_pt, lm_Left, (e.mouse.b & ss_Ctrl) != 0, (e.mouse.b & ss_Shift) != 0);
                } else if (e.mouse.b & mb_Right) {
                    doItemLock(new_pt, lm_Right, (e.mouse.b & ss_Ctrl) != 0, (e.mouse.b & ss_Shift) != 0);
                } else {
                    userMoveScannerTo(new_pt);
                }

                // Double-click exits
                if (e.mouse.b & mb_Double) {
                    running = false;
                }
            }
            break;
         case ev_Idle:
            break;
        }
    }
}

void
WScannerChartWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_Flexible;
    info.min_size = info.pref_size = GfxPoint(100,100);
}
