/**
  *  \file game/cmds.h
  *  \brief Auxiliary Command Management
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_CMDS_H
#define PCC_V2_GAME_CMDS_H

#include <list>

#include "cpluslib/string.h"

class GTimestamp;
class Stream;

class GCommand {
 public:
    /** Command type. */
    enum Type {
        phc_Language,           /**< Change language.
                                   - Id: zero
                                   - Arg: language name */
        phc_SendConfig,         /**< Request configuration.
                                    - Id: zero
                                    - Arg: ignored */
        phc_SendRacenames,      /**< Request race name file.
                                   - Id: zero
                                   - Arg: ignored */
        phc_SetRaceName,        /**< Change race name.
                                   - Id: phcrn_XXX
                                   - Arg: new name */
        phc_Filter,             /**< Configure message filter.
                                   - Id: zero
                                   - Arg: new status, "yes" or "no" */
        phc_ConfigAlly,         /**< Configure an alliance.
                                   - Id: player number
                                   - Arg: alliance status string ("+c ~m -v" etc.) */
        phc_AddDropAlly,        /**< Offer/Resign an alliance.
                                   - Id: player number
                                   - Arg: "add" or "drop" */
        phc_GiveShip,           /**< Give away a ship.
                                   - Id: ship Id
                                   - Arg: player number as string */
        phc_GivePlanet,         /**< Give away a planet.
                                   - Id: planet Id
                                   - Arg: player number as string */
        phc_RemoteControl,      /**< Configure remote control.
                                   - Id: ship Id
                                   - Arg: new status, "control", "allow", "forbid", "drop" */
        phc_RemoteDefault,      /**< Configure default remote control status.
                                   - Id: zero
                                   - Arg: new status, "allow" or "forbid" */
        phc_Beamup,             /**< Beam up multiple.
                                   - Id: ship Id
                                   - Arg: cargo string ("M100 Nmax") */
        phc_TAlliance,          /**< Configure THost alliances.
                                   - Id: zero
                                   - Arg: fcode list (3*N chars, "ffaFF1ee2") */
        phc_SendFcodes,         /**< Request friendly code list.
                                   - Id: zero
                                   - Arg: ignored */
        phc_SendFile,           /**< File relay (using tcm_SendBack, PHost 4.0+).
                                   - Id: zero
                                   - Arg: file name */
        phc_Enemies,            /**< Configure permanent enemies (PHost 4.0g+).
                                   - Id: player number
                                   - Arg: "add" or "drop" */
        phc_Unload,             /**< Unload stuff from ship to planet/space (PHost 4.0h+).
                                   - Id: ship Id
                                   - Arg: cargo string */
        phc_Transfer,           /**< Transfer stuff from ship to other ship (PHost 4.0h+).
                                   - Id: ship Id
                                   - Arg: cargo string plus "to TARGET-SID" */
        phc_ShowShip,           /**< Show ship to ally (PHost 4.0h+).
                                   - Id: ship Id
                                   - Arg: race list, "1 2 3" */
        phc_ShowPlanet,         /**< Show planet to ally (PHost 4.0h+).
                                   - Id: planet Id
                                   - Arg: race list, "1 2 3" */
        phc_ShowMinefield,      /**< Show minefield to ally (PHost 4.0h+).
                                   - Id: minefield Id
                                   - Arg: race list, "1 2 3" */
        phc_Refit,              /**< Extended refit order (PHost 4.0h+).
                                   - Id: ship Id
                                   - Arg: equipment */
        phc_Other,              /**< Other command.
                                   - Id: zero
                                   - Arg: the command */
        phc_LAST = phc_Other    /**< Identifies the last phc_XXX value. */
    };

    /** "id" parameters for phc_SetRaceName. */
    enum {
        phcrn_Long,
        phcrn_Short,
        phcrn_Adjective
    };

 private:
    Type      cmd;              ///< Command code.
    int       id;               ///< Associated Id.
    string_t  arg;              ///< Parameter.
 public:
    /** Create new command. */
    GCommand(Type cmd, int id, string_t arg)
        : cmd(cmd), id(id), arg(arg) { }

    ~GCommand();

    /** Get command code. */
    Type getCommand() const
        { return cmd; }
    /** Get associated Id number. */
    int getId() const
        { return id; }
    /** Get command parameter. */
    const string_t& getArg() const
        { return arg; }
    /** Set command parameter. */
    void setArg(const string_t& s)
        { arg = s; }

    int getAffectedShip() const;
    int getAffectedPlanet() const;
    int getAffectedMinefield() const;

    string_t getCommandText() const;

    static GCommand* parseCommand(string_t text, bool fromfile);
    static bool isMessageIntroducer(string_t text);
};

class GCommandContainer {
    typedef std::list<GCommand*> command_list;
    command_list cmds;
    command_list::iterator findCommand(GCommand::Type typ, int id);
    static int getValue(GCommand::Type type);
    void insertNewCommand(GCommand* cmd);
 public:
    GCommandContainer();
    ~GCommandContainer();

    void clear();

    const GCommand* getCommand(GCommand::Type typ, int id);
    const GCommand* setCommand(GCommand::Type typ, int id, string_t arg);
    const GCommand* setCommand(GCommand* cmd);
    bool            removeCommand(GCommand::Type typ, int id);
    void            removeCommand(GCommand* cmd);

    typedef command_list::iterator iterator;
    typedef command_list::const_iterator const_iterator;
    iterator begin()
        { return cmds.begin(); }
    const_iterator begin() const
        { return cmds.begin(); }
    iterator end()
        { return cmds.end(); }
    const_iterator end() const
        { return cmds.end(); }
    iterator erase(iterator it)
        { return cmds.erase(it); }

    void loadCommandFile(Stream& file, const GTimestamp& time);
};

#endif
