/**
  *  \file game/un-trnflt.cc
  *  \brief Implementation of Filter Classes for Un-TRN
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <cctype>
#include "game/un-trnflt.h"
#include "util/nls.h"
#include "io/io.h"
#include "util/recode.h"

/* ISO C++ mandates these are not macros, but libc5 does not care. */
#undef isspace
#undef isalpha
#undef isdigit

bool
FilterConst::accept(GTurnfile&, int)
{
    return res;
}

bool
FilterName::accept(GTurnfile& trn, int index)
{
    const char* cmname = trn.getCommandName(index);
    if (!cmname)
        return false;

    string_t s(cmname);
    if (wild && s.length() > str.length())
        s.erase(str.length());

    if (strCaseCmp(s, str) == 0)
        return true;
    else
        return false;
}

bool
FilterString::accept(GTurnfile& trn, int index)
{
    string_t text;
    switch(trn.getCommandCode(index)) {
     case tcm_ShipChangeFc:
     case tcm_PlanetChangeFc:
        text = convertGameToUtf8(getBasicStringN(trn.getCommandData(index), 3));
        break;
     case tcm_ShipChangeName:
        text = convertGameToUtf8(getBasicStringN(trn.getCommandData(index), 20)); // FIXME: recode
        break;
     case tcm_SendMessage:
        text.assign(trn.getCommandData(index), trn.getCommandId(index));
        for (string_t::size_type i = 0; i < text.length(); ++i)
            text[i] -= 13;
        text = convertGameToUtf8(text);
        break;
     default:
        return false;
    }
    return strUCase(text).find(str) != string_t::npos;
}

bool
FilterId::accept(GTurnfile& trn, int index)
{
    int type = trn.getCommandType(index);
    if (type == GTurnfile::ShipCommand
        || type == GTurnfile::PlanetCommand
        || type == GTurnfile::BaseCommand)
    {
        int id = trn.getCommandId(index);
        return (id >= lower) && (id <= upper);
    } else {
        return false;
    }
}

bool
FilterAnd::accept(GTurnfile& trn, int index)
{
    return lhs.accept(trn, index) && rhs.accept(trn, index);
}

bool
FilterOr::accept(GTurnfile& trn, int index)
{
    return lhs.accept(trn, index) || rhs.accept(trn, index);
}

bool
FilterNeg::accept(GTurnfile& trn, int index)
{
    return !other.accept(trn, index);
}

bool
FilterIndex::accept(GTurnfile& /*trn*/, int index)
{
    ++index;
    return (index >= from && index <= to);
}

/*
   filter ::= or
            | or "|" filter
   or     ::= and
            | and "&" or
   and    ::= '!' and
            | ex
   ex     ::= single
            | single ex
   single ::= "(" filter ")"
            | identifier
            | identifier "*"
            | number
            | number "-" number
            | "#" number
            | "#" number "-" number
            | "'" character* "'"
            | """ character* """
  */

static Filter& parseFilter1(const char*& text, ObjectHolder& h);

/** Skip leading whitespace of text. Advances text accordingly. */
static void
parseWs(const char*& text)
{
    while (*text && std::isspace(static_cast<unsigned char>(*text)))
        ++text;
}

static int
parseNumber(const char*& text)
{
    int result = 0;
    while (std::isdigit(static_cast<unsigned char>(*text)))
        result = 10*result + (*text++ - '0');
    return result;
}

/** Check whether text is the start of a 'single' production. */
static bool
isSingle(const char* text)
{
    unsigned char c = text[0];
    return c == '('
        || c == '\''
        || c == '"'
        || c == '#'
        || std::isalpha(c)
        || std::isdigit(c);
}

/*! \name UN-TRN Expression Parser */
//@{

/** Parse elementary expression.
    <pre>parseSingle := '(' parseFilter1 ')' | string | '#'? number ('-' number)? | identifier '*'?</pre> */
static Filter&
parseSingle(const char*& text, ObjectHolder& h)
{
    parseWs(text);
    if (*text == '(') {
        ++text;
        Filter& result = parseFilter1(text, h);
        parseWs(text);
        if (*text != ')')
            throw ParseException(_("`)' expected"));
        ++text;
        return result;
    } else if (*text == '\'' || *text == '"') {
        char delim = *text++;
        string_t value;
        while (*text && *text != delim)
            value += *text++;
        if (*text != delim)
            throw ParseException(_("unterminated string constant"));
        ++text;
        return h.add(new FilterString(value));
    } else if (std::isdigit(static_cast<unsigned char>(*text)) || *text == '#') {
        bool index = false;
        if (*text == '#') {
            index = true;
            ++text;
            parseWs(text);
            if (!std::isdigit(static_cast<unsigned char>(*text)))
                throw ParseException(_("Number expected after `#'"));
        }
        int low = parseNumber(text);
        int high = low;
        parseWs(text);
        if (*text == '-') {
            ++text;
            parseWs(text);
            if (!std::isdigit(static_cast<unsigned char>(*text)))
                throw ParseException(_("Number expected after `-'"));
            high = parseNumber(text);
        }
        if (index)
            return h.add(new FilterIndex(low, high));
        else
            return h.add(new FilterId(low, high));
    } else if (std::isalpha(static_cast<unsigned char>(*text))) {
        int i = 0;
        while (std::isalpha(static_cast<unsigned char>(text[i])))
            ++i;
        string_t value(text, i);
        text += i;
        parseWs(text);
        if (strCaseCmp(value, "true") == 0) {
            return h.add(new FilterConst(true));
        } else if (strCaseCmp(value, "false") == 0) {
            return h.add(new FilterConst(false));
        } else {
            if (*text == '*') {
                ++text;
                return h.add(new FilterName(value, true));
            } else {
                return h.add(new FilterName(value));
            }
        }
    } else
        throw ParseException(_("Invalid expression"));
}

/** Parse almost-elementary expression.
    <pre>parseEx ::= parseSingle+</pre> */
static Filter&
parseEx(const char*& text, ObjectHolder& h)
{
    Filter* result = &parseSingle(text, h);
    while (1) {
        parseWs(text);
        if (!isSingle(text))
            break;
        result = &h.add(new FilterAnd(*result, parseSingle(text, h)));
    }
    return *result;
}

/** Parse "and" expression.
    <pre>parseAnd ::= '!'? parseEx</pre> */
static Filter&
parseAnd(const char*& text, ObjectHolder& h)
{
    parseWs(text);
    if (*text == '!') {
        ++text;
        return h.add(new FilterNeg(parseAnd(text, h)));
    } else
        return parseEx(text, h);
}

/** Parse "or" expression.
    <pre>parseOr ::= parseAnd ('&' parseAnd)*</pre> */
static Filter&
parseOr(const char*& text, ObjectHolder& h)
{
    Filter* result = &parseAnd(text, h);
    while (1) {
        parseWs(text);
        if (*text != '&')
            break;
        ++text;
        result = &h.add(new FilterAnd(*result, parseAnd(text, h)));
    }
    return *result;
}

/** Parse filter expression, back-end.
    <pre>parseFilter1 ::= parseOr ('|' parseOr)*</pre> */
static Filter&
parseFilter1(const char*& text, ObjectHolder& h)
{
    Filter* result = &parseOr(text, h);
    while (1) {
        parseWs(text);
        if (*text != '|')
            break;
        ++text;
        result = &h.add(new FilterOr(*result, parseOr(text, h)));
    }
    return *result;
}

/** Parse filter expression.
    \param text text to parse
    \param h memory manager
    \returns filter expression
    \throws ParseException on error; exception object contains
    error message. */
Filter&
parseFilter(const char* text, ObjectHolder& h)
{
    Filter& result = parseFilter1(text, h);
    parseWs(text);
    if (*text)
        throw ParseException(_("Expression incorrectly terminated"));
    return result;
}

//@}
