/**
  *  \file util/pluginmgr.cc
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "util/pluginmgr.h"
#include "version.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/except.h"
#include "cpluslib/format.h"

static bool
comparePlugins(Plugin* a, Plugin* b)
{
    return a->getId() < b->getId();
}

static void
initFeatures(Plugin::FeatureSet& features)
{
    features["PCC"] = PCC2_VERSION;
}

PluginManager::PluginManager()
    : plugins()
{ }

PluginManager::~PluginManager()
{ }

PluginManager&
PluginManager::getInstance()
{
    static PluginManager mgr;
    return mgr;
}

void
PluginManager::findPlugins(Directory& dir)
{
    // Open the directory. This might throw, but is not a problem;
    // it just means the directory does not exist.
    Ptr<Enumeration<Ptr<DirectoryEntry> > > entries;
    try {
        entries = dir.getEntries();
    }
    catch (...) { }
    if (!entries.ptr()) {
        return;
    }

    try {
        // Load everything
        while (entries->hasMoreElements()) {
            Ptr<DirectoryEntry> elem = entries->getNextElement();
            string_t name = elem->getBaseName();
            if (name.size() > 4
                && strCaseCmp(name.substr(name.size() - 4), ".c2p") == 0
                && name[0] != '.'
                && elem->getFileType() == DirectoryEntry::FT_FILE)
            {
                // Looks like a plugin
                string_t pluginName = name.substr(0, name.size() - 4);
                try {
                    std::auto_ptr<Plugin> plugin(new Plugin(strUCase(pluginName)));
                    Ptr<Stream> file = elem->openFile(Stream::C_READ);
                    plugin->initFromPluginFile(dir.getEntryByName(pluginName)->getName(),
                                               name,
                                               *file);
                    plugins.push_back(plugin.release());
                }
                catch (std::exception& e) {
                    console.write(LOG_ERROR, format(_("Error loading plugin %s: %s"),
                                                    pluginName,
                                                    e.what()));
                }
            }
        }

        // Sort for determinism
        plugins.sort(comparePlugins);
    }
    catch (FileProblemException& e) {
        console.write(LOG_ERROR, format(_("Cannot load plugins: %s"),
                                        e.getFileName() + ": " + e.what()));
    }
    catch (std::exception& e) {
        console.write(LOG_ERROR, format(_("Cannot load plugins: %s"), e.what()));
    }
}

void
PluginManager::findPlugins(string_t dirName)
{
    try {
        Ptr<Directory> dir = Directory::openDir(dirName);
        findPlugins(*dir);
    }
    catch (...) { }
}

void
PluginManager::addNewPlugin(Plugin* p)
{
    plugins.push_back(p);
}

/** Enumerate all plugins in dependency order. */
void
PluginManager::enumPlugins(std::vector<Plugin*>& out, bool ordered) const
{
    // Initial feature set
    Plugin::FeatureSet features;
    initFeatures(features);

    // Marker for all plugins
    std::vector<char> did;
    did.resize(plugins.size());

    // Enumerate
    while (1) {
        bool didOne = false;
        for (size_t i = 0, n = plugins.size(); i < n; ++i) {
            if (!did[i] && (!ordered || plugins[i]->isSatisfied(features))) {
                didOne = true;
                did[i] = 1;
                out.push_back(plugins[i]);
                plugins[i]->addProvidedFeaturesTo(features);
            }
        }
        if (!didOne) {
            break;
        }
    }

    // Anything missing?
    if (out.size() != plugins.size()) {
        for (size_t i = 0, n = plugins.size(); i < n; ++i) {
            if (!did[i]) {
                console.write(LOG_ERROR, format(_("Plugin %s cannot be loaded because of missing preconditions."), plugins[i]->getId()));
            }
        }
    }
}

/** Enumerate all plugins that conflict with the given candidate. */
void
PluginManager::enumConflictingPlugins(const Plugin& candidate, std::vector<Plugin*>& out) const
{
    for (size_t i = 0, n = plugins.size(); i < n; ++i) {
        if (candidate.getId() == plugins[i]->getId()
            ? !candidate.isUpdateFor(*plugins[i])
            : candidate.isConflict(*plugins[i]))
        {
            out.push_back(plugins[i]);
        }
    }
}

/** Enumerate all plugins that depend on the candidate plugin, i.e.\ prevent its uninstallation. */
void
PluginManager::enumDependingPlugins(const Plugin& candidate, std::vector<Plugin*>& out) const
{
    for (size_t i = 0, n = plugins.size(); i < n; ++i) {
        if (plugins[i]->isDependingOn(candidate)) {
            out.push_back(plugins[i]);
        }
    }
}

/** Enumerate all available features. */
void
PluginManager::enumFeatures(Plugin::FeatureSet& fset) const
{
    initFeatures(fset);
    for (size_t i = 0, n = plugins.size(); i < n; ++i) {
        plugins[i]->addProvidedFeaturesTo(fset);
    }
}

/** Extract plugin, given a pointer.
    Caller must delete it.
    \return pointer on success, otherwise 0 */
Plugin*
PluginManager::extractPlugin(Plugin* p)
{
    for (size_t i = 0, n = plugins.size(); i < n; ++i) {
        if (p == plugins[i]) {
            Plugin* retVal = plugins.extract_elem(i);
            plugins.erase(plugins.begin() + i);
            return retVal;
        }
    }
    return 0;
}

/** Get plugin, given an Id. */
Plugin*
PluginManager::getPluginById(const string_t& id) const
{
    for (size_t i = 0, n = plugins.size(); i < n; ++i) {
        if (plugins[i]->getId() == id) {
            return plugins[i];
        }
    }
    return 0;
}

